package com.example.aspdemo;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.res.Resources;
import android.os.Handler;
import android.os.Looper;
import android.util.Log;
import android.view.LayoutInflater;

import java.util.HashMap;

/**
 * A helper class used to manipulate dialog used to prompt user about info/warning/error raised from
 * ASPEngine implementation
 *
 * @author suhang.sh
 */
public class DialogHelper {
    private static final String TAG = "DialogHelper";

    private static DialogHelper mInstance = null;
    private static Object mLocker = new Object();

    private AlertDialog.Builder mBuilder = null;
    private StateBase mCurrentState = null;
    private StateBase mLastState = null;
    private Context mContext = null;
    private Resources mResources = null;
    private LayoutInflater mInflater = null;
    private Handler mHandler = new Handler(Looper.getMainLooper());
    private HashMap<State, StateBase> mStateMaps = new HashMap<State, StateBase>();

    public enum State {
        STATE_CONNECTING,
        STATE_CONNECTED,
        STATE_DISCONNECTING,
        STATE_DISCONNECTED,
        STATE_ENGINE_ERROR,
        STATE_CONNECTION_ERROR,
    }

    abstract class StateBase {
        protected AlertDialog mDialog = null;

        protected Runnable mDismissRunnable = new Runnable() {
            @Override
            public void run() {
                Log.i(TAG, "Dismiss dialog " + mDialog + " for " + StateBase.this);
                if (mDialog != null) {
                    mDialog.dismiss();
                }
            }
        };

        class EnterStateRunnable implements Runnable {
            private String mMsg1 = null;
            private String mMsg2 = null;

            EnterStateRunnable(String msg1, String msg2) {
                mMsg1 = msg1;
                mMsg2 = msg2;
            }

            @Override
            public void run() {
                Log.i(TAG, "last state " + mLastState);
                Log.i(TAG, "current state " + StateBase.this);

                if (mBuilder == null) {
                    return;
                }

                if (mDialog == null) {
                    mBuilder.setTitle(getTitle());
                    mBuilder.setMessage(getMessage(mMsg1, mMsg2));
                    //mBuilder.setView(getContentView(msg1, msg2));

                    boolean cancellable = isCancellable();
                    mBuilder.setCancelable(cancellable);

                    mDialog = mBuilder.create();

                    setButtons(mDialog);

                    mDialog.setCanceledOnTouchOutside(cancellable);
                }

                mDialog.show();

                if (getDismissTimeout() > 0) {
                    Log.i(TAG, "Dismiss dialog for state " + mCurrentState + " after "
                            + getDismissTimeout() + " millis");
                    runTaskOnMainThread(mDismissRunnable, getDismissTimeout());
                }
            }
        }
        protected EnterStateRunnable mEnterStateRunnable = null;

        boolean enter() {
            return enter(null, null);
        }

        boolean enter(String msg1, String msg2) {
            if (canEnter()) {
                if (mLastState != null) {
                    mLastState.onLeave();
                }

                Log.i(TAG, "enter state " + StateBase.this);

                if (mEnterStateRunnable == null) {
                    mEnterStateRunnable = new EnterStateRunnable(msg1, msg2);
                }
                runTaskOnMainThread(mEnterStateRunnable);
                return true;
            }
            return false;
        }

        protected abstract boolean canEnter();
        protected abstract String getTitle();
        protected abstract String getMessage(String msg1, String msg2);
        protected abstract void setButtons(AlertDialog dialog);
        protected abstract boolean isCancellable();

        protected long getDismissTimeout() {
            return -1;
        }

        protected void onLeave() {
            Log.i(TAG, "onLeave " + this);
            mHandler.removeCallbacks(mEnterStateRunnable);
            mHandler.removeCallbacks(mDismissRunnable);
            runTaskOnMainThread(mDismissRunnable);
        }

        private void runTaskOnMainThread(Runnable r) {
            runTaskOnMainThread(r, -1);
        }

        private boolean isMainThread() {
            return Looper.getMainLooper() == Looper.myLooper();
        }

        private void runTaskOnMainThread(Runnable r, long delayMS) {
            if (isMainThread() && delayMS <= 0) {
                r.run();
            } else if (delayMS <= 0) {
                mHandler.post(r);
            } else {
                mHandler.postDelayed(r, delayMS);
            }
        }
    }

    class ConnectingState extends StateBase {
        @Override
        protected boolean canEnter() {
            return true;
        }

        @Override
        protected String getTitle() {
            return null;
        }

        @Override
        protected String getMessage(String msg1, String msg2) {
            return mResources.getString(R.string.dialog_helper_connecting_content);
        }

        @Override
        protected void setButtons(AlertDialog dialog) {
            // Do nothing
        }

        @Override
        protected boolean isCancellable() {
            return false;
        }
    }

    class ConnectedState extends StateBase {
        final long DISMISS_DELAY_MS = 2000;

        @Override
        protected boolean canEnter() {
            return true;
        }

        @Override
        protected String getTitle() {
            return null;
        }

        @Override
        protected String getMessage(String msg1, String msg2) {
            return mResources.getString(R.string.dialog_helper_connected_content);
        }

        @Override
        protected void setButtons(AlertDialog dialog) {
            // Do nothing
        }

        @Override
        protected boolean isCancellable() {
            return true;
        }

        @Override
        protected long getDismissTimeout() {
            return DISMISS_DELAY_MS;
        }
    }

    class DisconnectingState extends StateBase {
        @Override
        protected boolean canEnter() {
            return true;
        }

        @Override
        protected String getTitle() {
            return null;
        }

        @Override
        protected String getMessage(String msg1, String msg2) {
            return mResources.getString(R.string.dialog_helper_disconnecting_content);
        }

        @Override
        protected void setButtons(AlertDialog dialog) {

        }

        @Override
        protected boolean isCancellable() {
            return false;
        }
    }

    class DisconnectedState extends StateBase {
        @Override
        protected boolean canEnter() {
            return true;
        }

        @Override
        protected String getTitle() {
            return null;
        }

        @Override
        protected String getMessage(String msg1, String msg2) {
            return mResources.getString(R.string.dialog_helper_disconnected_content);
        }

        @Override
        protected void setButtons(AlertDialog dialog) {
            dialog.setButton(AlertDialog.BUTTON_POSITIVE,
                    mResources.getText(R.string.dialog_helper_button_ok),
                    new DialogInterface.OnClickListener() {
                        @Override
                        public void onClick(DialogInterface dialogInterface, int i) {
                            dialog.dismiss();
                            if (mContext instanceof Activity) {
                                Activity activity = (Activity)mContext;
                                activity.finish();
                            }
                        }
                    });
        }

        @Override
        protected boolean isCancellable() {
            return false;
        }
    }

    class EngineErrorState extends StateBase {
        @Override
        protected boolean canEnter() {
            return true;
        }

        @Override
        protected String getTitle() {
            return mResources.getString(R.string.dialog_helper_engine_error_title);
        }

        @Override
        protected String getMessage(String msg1, String msg2) {
            String errorCode = mResources.getString(R.string.dialog_helper_error_code, msg1);
            String errorMessage = mResources.getString(R.string.dialog_helper_error_message, msg2);
            return errorCode + ", " + errorMessage;
        }

        @Override
        protected void setButtons(AlertDialog dialog) {
            dialog.setButton(AlertDialog.BUTTON_POSITIVE,
                    mResources.getText(R.string.dialog_helper_button_ok),
                    new DialogInterface.OnClickListener() {
                        @Override
                        public void onClick(DialogInterface dialogInterface, int i) {
                            dialog.dismiss();
                        }
                    });
        }

        @Override
        protected boolean isCancellable() {
            return false;
        }
    }

    class ConnectionErrorState extends StateBase {
        @Override
        protected boolean canEnter() {
            return true;
        }

        @Override
        protected String getTitle() {
            return mResources.getString(R.string.dialog_helper_connection_error_title);
        }

        @Override
        protected String getMessage(String msg1, String msg2) {
            String errorCode = mResources.getString(R.string.dialog_helper_error_code, msg1);
            String errorMessage = mResources.getString(R.string.dialog_helper_error_message, msg2);
            return errorCode + ", " + errorMessage;
        }

        @Override
        protected void setButtons(AlertDialog dialog) {
            dialog.setButton(AlertDialog.BUTTON_POSITIVE,
                    mResources.getText(R.string.dialog_helper_button_ok),
                    new DialogInterface.OnClickListener() {
                        @Override
                        public void onClick(DialogInterface dialogInterface, int i) {
                            dialog.dismiss();
                            if (mContext instanceof Activity) {
                                Activity activity = (Activity)mContext;
                                activity.finish();
                            }
                        }
                    });
        }

        @Override
        protected boolean isCancellable() {
            return false;
        }
    }

    private DialogHelper() {

    }

    public static DialogHelper getInstance() {
        synchronized (mLocker) {
            if (mInstance == null) {
                mInstance = new DialogHelper();
            }
        }
        return mInstance;
    }

    public synchronized void initialize(Context context) {
        reset();
        mBuilder = new AlertDialog.Builder(context);
        mContext = context;
        mInflater = LayoutInflater.from(context);
        mResources = mContext.getResources();
        mCurrentState = mLastState = null;
        Log.i(TAG, "initialize done, context " + context
                + " inflater " + mInflater
                + " resources " + mResources);
    }

    public synchronized void reset() {
        Log.i(TAG, "reset");
        mBuilder = null;
        mContext = null;
        mInflater = null;
        mResources = null;
        mStateMaps.clear();
        if (mCurrentState != null) {
            mCurrentState.onLeave();
        }
        if (mLastState != null) {
            mLastState.onLeave();
        }
        mCurrentState = mLastState = null;
    }

    public void setState(State state) {
        setState(state, null, null);
    }

    public synchronized void setState(State state, String msg1, String msg2) {
        Log.i(TAG, "To state " + state);

        mLastState = mCurrentState == null ? mLastState : mCurrentState;
        mCurrentState = getState(state);

        if (mCurrentState == null) {
            Log.e(TAG, "Unknown state!" + state);
            return;
        }

        if (msg1 != null) {
            mCurrentState.enter(msg1, msg2);
        } else {
            mCurrentState.enter();
        }
    }

    private StateBase getState(State state) {
        if (mStateMaps.containsKey(state)) {
            return mStateMaps.get(state);
        }

        StateBase stateBase = null;
        switch (state) {
            case STATE_CONNECTING:
                stateBase = new ConnectingState();
                break;
            case STATE_CONNECTED:
                stateBase = new ConnectedState();
                break;
            case STATE_DISCONNECTING:
                stateBase = new DisconnectingState();
                break;
            case STATE_DISCONNECTED:
                stateBase = new DisconnectedState();
                break;
            case STATE_ENGINE_ERROR:
                stateBase = new EngineErrorState();
                break;
            case STATE_CONNECTION_ERROR:
                stateBase = new ConnectionErrorState();
                break;
            default:
                return null;
        }
        mStateMaps.put(state, stateBase);
        return stateBase;
    }
}
