package com.example.aspdemo.mediastreamplayer.video;

import android.media.MediaCodec;
import android.media.MediaFormat;
import android.os.Build;
import android.util.Log;
import android.view.Surface;

import com.aliyun.wuying.aspsdk.aspengine.VCodecType;
import com.aliyun.wuying.aspsdk.aspengine.VFrame;
import com.aliyun.wuying.aspsdk.aspengine.VProfile;

import java.nio.ByteBuffer;

class MediaCodecVideoDecoder {
    private static final String TAG = "MediaCodecVideoDecoder";

    private final long DEQUEUE_BUFFER_TIMEOUT = 10000;
    private MediaCodec mDecoder = null;
    private MediaFormat mFormat = null;
    private Thread mDecoderOutputHandler = null;
    private boolean mOutputThreadRunning = false;
    private boolean mConfigured = false;

    class DecoderOutputHandler implements Runnable {
        @Override
        public void run() {
            while (mOutputThreadRunning) {
                if (mDecoder != null) {
                    MediaCodec.BufferInfo info = new MediaCodec.BufferInfo();
                    int outIndex = mDecoder.dequeueOutputBuffer(info, DEQUEUE_BUFFER_TIMEOUT);
                    if (outIndex >= 0) {
                        if ((info.flags & MediaCodec.BUFFER_FLAG_END_OF_STREAM) != 0) {
                            Log.i(TAG, "output EOS");
                        }
                        mDecoder.releaseOutputBuffer(outIndex, info.size != 0);
                    }
                }
            }
        }
    }

    public boolean setVideoProfile(VProfile profile) {
        if (profile.codecType == VCodecType.H264) {
            mFormat = new MediaFormat();
            mFormat.setString(MediaFormat.KEY_MIME, "video/avc");
            mFormat.setInteger(MediaFormat.KEY_WIDTH, profile.width);
            mFormat.setInteger(MediaFormat.KEY_HEIGHT, profile.height);

            try {
                mDecoder = MediaCodec.createDecoderByType("video/avc");
            } catch (Exception ex) {
                Log.e(TAG, "failed to create video decoder");
                return false;
            }
        }

        return true;
    }

    public boolean setSurface(Surface surface) {
        if (!mConfigured && surface != null) {
            mDecoder.configure(mFormat, surface, null, 0);
            mConfigured = true;
        } else if (mConfigured) {
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                mDecoder.setOutputSurface(surface);
            } else {
                // Otherwise, need to rebuild the mediacodec
            }
        }

        return true;
    }

    public boolean start() {
        if (mDecoder != null && !mOutputThreadRunning) {
            mDecoder.start();
            mOutputThreadRunning = true;
            mDecoderOutputHandler = new Thread(new DecoderOutputHandler());
            mDecoderOutputHandler.start();
        }

        return true;
    }

    public boolean stop() {
        if (mDecoder != null) {
            mOutputThreadRunning = false;
            try {
                mDecoderOutputHandler.join();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            mDecoder.stop();
            mDecoder = null;
            mConfigured = false;
        }

        return true;
    }

    public boolean handleInputFrame(VFrame frame) {
        if (mDecoder != null) {
            int tryTimes = 0;
            do {
                int index = mDecoder.dequeueInputBuffer(DEQUEUE_BUFFER_TIMEOUT);
                if (index >= 0) {
                    ByteBuffer buffer = mDecoder.getInputBuffer(index);
                    if (buffer == null) {
                        Log.e(TAG, "No available codec buffer, index " + index);
                        return false;
                    } else {
                        buffer.put(frame.frame);
                        mDecoder.queueInputBuffer(index, 0, frame.size, frame.mmTime, 0);
                        Log.d(TAG, "buffer queued");
                        break;
                    }
                } else {
                    try {
                        Thread.sleep(100);
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            } while (tryTimes++ < 5);
        }

        return true;
    }
}
